use super::{
  directory::{index::IndexStrategy, Directory},
  named::NamedFile,
  pages::render_server_page,
  path::resolve_path_within_tree,
  path_context::PathContext,
  pathbuf::UriPathBuf,
};
use actix_service::{boxed::BoxService, Service};
use actix_web::{
  dev::{ServiceRequest, ServiceResponse},
  error::Error as ActixError,
  guard::Guard,
  http::{header, Method, StatusCode},
  HttpResponse, ResponseError,
};
use futures_util::future::{err, ok, Either, LocalBoxFuture, Ready};
use snafu::ResultExt;
use snafu::Snafu;
use std::{
  convert::TryInto,
  io,
  path::PathBuf,
  rc::Rc,
  sync::{Arc, RwLock},
  task::{Context, Poll},
};

/// Errors which can occur when serving static files.
#[derive(Snafu, Debug)]
pub enum Error {
  /// Path is not a directory
  #[snafu(display("Path is not a directory. Unable to serve static files"))]
  IsNotDirectory,

  /// Cannot render directory
  #[snafu(display("Unable to render directory without index file"))]
  IsDirectory,

  #[snafu(display("Serve directory is not ready."))]
  ServeDirNotReady,

  #[snafu(display("Unable to obtain read lock for serve dir."))]
  ServerDirReadLockFail,

  #[snafu(display("Unable to find or open the specified file."))]
  OpenFileError {
    source: io::Error,
  },

  CanonicalizePath {
    source: io::Error,
  },

  MethodNotAllowed {
    method: Method,
  },

  BuildUriPath {
    source: super::pathbuf::Error,
  },

  RenderListing {
    source: io::Error,
  },

  ReconstructRequest,

  BuildFileResponse {
    source: ActixError,
  },

  ServeCustomNotFoundPath,
}

/// Return `NotFound` for `Error`
impl ResponseError for Error {
  fn error_response(&self) -> HttpResponse {
    match self {
      Error::MethodNotAllowed { .. } => HttpResponse::MethodNotAllowed()
        .header(header::CONTENT_TYPE, "text/plain")
        .body("Request did not meet this resource's requirements."),
      _ => HttpResponse::new(StatusCode::NOT_FOUND),
    }
  }
}

type HttpService = BoxService<ServiceRequest, ServiceResponse, ActixError>;

struct RequestContext {
  serve_dir: PathBuf,
  path: PathBuf,
  path_context: PathContext,
}

pub struct FilesService {
  directory: Arc<RwLock<Option<PathBuf>>>,
  redirect_to_slash: bool,
  default: Option<HttpService>,
  root_path_context: Arc<PathContext>,
  #[allow(clippy::rc_buffer)]
  path_contexts: Arc<Vec<PathContext>>,
  // FIXME: Should re-visit later.
  #[allow(clippy::redundant_allocation)]
  guards: Option<Rc<dyn Guard>>,
}

impl FilesService {
  #[allow(clippy::rc_buffer)]
  pub(crate) fn new(
    directory: Arc<RwLock<Option<PathBuf>>>,
    redirect_to_slash: bool,
    default: Option<HttpService>,
    root_path_context: Arc<PathContext>,
    path_contexts: Arc<Vec<PathContext>>,
    guards: Option<Rc<dyn Guard>>,
  ) -> FilesService {
    FilesService {
      directory,
      redirect_to_slash,
      default,
      root_path_context,
      path_contexts,
      guards,
    }
  }

  pub fn set_default(&mut self, default: Option<HttpService>) {
    self.default = default
  }

  fn handle_err(
    &mut self,
    e: Error,
    req: ServiceRequest,
    request_context: Option<RequestContext>,
  ) -> Either<
    Ready<Result<ServiceResponse, ActixError>>,
    LocalBoxFuture<'static, Result<ServiceResponse, ActixError>>,
  > {
    log::debug!("Files: Failed to handle {}: {}", req.path(), e);

    if let Some(ref mut default) = self.default {
      info!("Using default handler");

      Either::Right(default.call(req))
    } else {
      let error_response = e.error_response();

      match error_response.status() {
        StatusCode::NOT_FOUND => self.handle_not_found_err(req, request_context),
        _ => Either::Left(ok(req.into_response(error_response))),
      }
    }
  }

  fn handle_early_err(
    &mut self,
    e: Error,
    req: ServiceRequest,
  ) -> Either<
    Ready<Result<ServiceResponse, ActixError>>,
    LocalBoxFuture<'static, Result<ServiceResponse, ActixError>>,
  > {
    let request_context: Result<RequestContext, Error> = try {
      let serve_dir = self.get_serve_dir()?;
      let path_from_request: UriPathBuf = (&req).try_into().context(BuildUriPath)?;

      RequestContext {
        path: serve_dir.join(&path_from_request),
        serve_dir,
        path_context: self.get_path_context(&path_from_request.into()),
      }
    };

    self.handle_err(e, req, request_context.ok())
  }

  fn handle_not_found_err(
    &mut self,
    req: ServiceRequest,
    request_context: Option<RequestContext>,
  ) -> Either<
    Ready<Result<ServiceResponse, ActixError>>,
    LocalBoxFuture<'static, Result<ServiceResponse, ActixError>>,
  > {
    if let Some(request_context) = request_context {
      debug!(
        "Using custom error path: {:?}",
        request_context.path_context.get_not_found_path()
      );

      // Try to use a custom 404 page, if configured.
      if let Some(not_found_path) = request_context.path_context.get_not_found_path() {
        match resolve_path_within_tree(
          &request_context.serve_dir,
          &request_context.path,
          not_found_path,
        ) {
          Ok(path) => {
            debug!("Custom error page: {:?}", path);

            return self.handle_file(
              req,
              RequestContext {
                path,
                ..request_context
              },
              false,
            );
          }
          Err(err) => error!("Error resolving custom error page: {}", err),
        }
      }
    }

    match render_server_page("404", "Not Found", "") {
      Ok(body) => Either::Left(ok(
        req.into_response(
          HttpResponse::NotFound()
            .content_type("text/html; charset=utf-8")
            .body(&body),
        ),
      )),
      Err(_) => self.handle_err(Error::ServeCustomNotFoundPath, req, None),
    }
  }

  fn handle_dir(
    &mut self,
    req: ServiceRequest,
    request_context: RequestContext,
  ) -> Either<
    Ready<Result<ServiceResponse, ActixError>>,
    LocalBoxFuture<'static, Result<ServiceResponse, ActixError>>,
  > {
    // Check if we should redirect the client to a normalized path ending in a
    // slash.
    if self.redirect_to_slash && !req.path().ends_with('/') {
      let redirect_to = format!("{}/", req.path());

      return Either::Left(ok(
        req.into_response(
          HttpResponse::Found()
            .header(header::LOCATION, redirect_to)
            .body("")
            .into_body(),
        ),
      ));
    }

    let RequestContext {
      path,
      path_context,
      serve_dir,
    } = &request_context;

    return match path_context.get_index_strategy_or_default() {
      IndexStrategy::AlwaysShowListing { renderer } => {
        let dir = Directory::new(serve_dir.clone(), path.clone());
        let (http_req, payload) = req.into_parts();

        match renderer(&dir, &http_req) {
          Ok(resp) => Either::Left(ok(resp)),
          Err(source) => match ServiceRequest::from_parts(http_req, payload) {
            Ok(req) => self.handle_err(Error::RenderListing { source }, req, Some(request_context)),
            Err(_) => Either::Left(err(Error::ReconstructRequest.into())),
          },
        }
      }
      IndexStrategy::ShowListingWhenAbsent {
        renderer,
        filenames,
      } => {
        for filename in filenames {
          let path = path.join(filename);

          if path.exists() {
            return self.handle_file(
              req,
              RequestContext {
                serve_dir: request_context.serve_dir,
                path,
                path_context: request_context.path_context,
              },
              true,
            );
          }
        }

        let dir = Directory::new(serve_dir.clone(), path.clone());
        let (http_req, payload) = req.into_parts();

        match renderer(&dir, &http_req) {
          Ok(resp) => Either::Left(ok(resp)),
          Err(source) => match ServiceRequest::from_parts(http_req, payload) {
            Ok(req) => self.handle_err(Error::RenderListing { source }, req, Some(request_context)),
            Err(_) => Either::Left(err(Error::ReconstructRequest.into())),
          },
        }
      }
      IndexStrategy::IndexFiles { filenames } => {
        for filename in filenames {
          let path = path.join(filename);

          if path.exists() {
            return self.handle_file(
              req,
              RequestContext {
                path,
                ..request_context
              },
              true,
            );
          }
        }

        self.handle_err(Error::IsDirectory, req, Some(request_context))
      }
      IndexStrategy::NoIndex => self.handle_err(Error::IsDirectory, req, Some(request_context)),
    };
  }

  fn handle_file(
    &mut self,
    req: ServiceRequest,
    request_context: RequestContext,
    preserve_request_context: bool,
  ) -> Either<
    Ready<Result<ServiceResponse, ActixError>>,
    LocalBoxFuture<'static, Result<ServiceResponse, ActixError>>,
  > {
    let RequestContext {
      path, path_context, ..
    } = &request_context;

    match NamedFile::open(path).context(OpenFileError) {
      Ok(named_file) => {
        let (http_req, payload) = req.into_parts();

        match named_file
          .with_path_context(&path_context)
          .into_response(&http_req)
        {
          Ok(mut response) => {
            path_context.append_headers_to_map(response.headers_mut());

            Either::Left(ok(ServiceResponse::new(http_req, response)))
          }
          Err(source) => match ServiceRequest::from_parts(http_req, payload) {
            Ok(req) => self.handle_err(
              Error::BuildFileResponse { source },
              req,
              if preserve_request_context {
                Some(request_context)
              } else {
                None
              },
            ),
            Err(_) => Either::Left(err(Error::ReconstructRequest.into())),
          },
        }
      }
      Err(e) => self.handle_err(
        e,
        req,
        if preserve_request_context {
          Some(request_context)
        } else {
          None
        },
      ),
    }
  }

  fn get_serve_dir(&self) -> Result<PathBuf, Error> {
    let maybe_dir = match self.directory.read().map(|dir| dir.clone()) {
      Ok(maybe_dir) => maybe_dir,
      Err(_) => return Err(Error::ServerDirReadLockFail),
    };

    match maybe_dir {
      Some(dir) => Ok(dir),
      None => Err(Error::ServeDirNotReady),
    }
  }

  fn get_canonical_path_from_request(
    &self,
    req: &ServiceRequest,
  ) -> Result<(PathBuf, PathBuf, UriPathBuf), Error> {
    let serve_dir = self.get_serve_dir()?;

    let path_from_request: UriPathBuf = req.try_into().context(BuildUriPath)?;

    // Obtain the normalized full file path to use.
    let canonical_path = serve_dir
      .join(&path_from_request)
      .canonicalize()
      .context(CanonicalizePath)?;

    Ok((canonical_path, serve_dir, path_from_request))
  }

  fn get_path_context(&self, path_from_request: &PathBuf) -> PathContext {
    for path_context in self.path_contexts.as_ref() {
      if path_context.matches_path(path_from_request) {
        debug!(
          "Using matched PathContext. Matcher: {:?}, Path: {}",
          path_context,
          path_from_request.display()
        );
        return self.root_path_context.merge(path_context);
      }
    }

    debug!(
      "Using root PathContext. Path: {}",
      path_from_request.display()
    );
    self.root_path_context.as_ref().clone()
  }
}

impl Service for FilesService {
  type Request = ServiceRequest;
  type Response = ServiceResponse;
  type Error = ActixError;
  type Future = Either<
    Ready<Result<Self::Response, Self::Error>>,
    LocalBoxFuture<'static, Result<Self::Response, Self::Error>>,
  >;

  fn poll_ready(&mut self, _: &mut Context) -> Poll<Result<(), Self::Error>> {
    Poll::Ready(Ok(()))
  }

  fn call(&mut self, req: ServiceRequest) -> Self::Future {
    // Check user-defined guards or fallback to default method check.
    let is_method_valid = if let Some(guard) = &self.guards {
      (**guard).check(req.head())
    } else {
      matches!(*req.method(), Method::HEAD | Method::GET)
    };

    if !is_method_valid {
      return self.handle_early_err(
        Error::MethodNotAllowed {
          method: req.method().clone(),
        },
        req,
      );
    }

    let maybe_path = self.get_canonical_path_from_request(&req);
    let (path, serve_dir, path_from_request) = match maybe_path {
      Ok(path) => path,
      Err(e) => return self.handle_early_err(e, req),
    };

    let path_context = self.get_path_context(&path_from_request.into());

    let request_context = RequestContext {
      path,
      serve_dir,
      path_context,
    };

    if request_context.path.is_dir() {
      self.handle_dir(req, request_context)
    } else {
      self.handle_file(req, request_context, true)
    }
  }
}
