use crate::{
  config::{CompressionConfig, ServerConfig},
  files::{path_context::PathContext, Files},
  thread,
  thread::ThreadHandle,
};
use actix::System;
use actix_http::http::{uri::InvalidUri, ContentEncoding};
use actix_web::{
  dev::Server as ActixServer,
  middleware::{self, Logger},
  App, HttpServer,
};
use snafu::{ResultExt, Snafu};
use std::{
  convert::{TryFrom, TryInto},
  path::PathBuf,
  sync::{
    mpsc::{self, RecvError, SendError, Sender},
    Arc, RwLock,
  },
  thread::{JoinHandle, Thread},
};

#[derive(Debug, Snafu)]
pub enum Error {
  ChannelSend {
    source: SendError<Server>,
  },
  ChannelReceive {
    source: RecvError,
  },
  Bind {
    source: std::io::Error,
  },
  SystemRun {
    source: std::io::Error,
  },
  RedirectUrlParseError {
    source: InvalidUri,
  },
  CreatePathContext {
    source: crate::files::path_context::Error,
  },
}

pub type Result<T, E = Error> = std::result::Result<T, E>;

pub struct Server {
  config: ServerConfig,
  serve_dir: Arc<RwLock<Option<PathBuf>>>,
}

impl Server {
  pub fn new(config: ServerConfig, serve_dir: Arc<RwLock<Option<PathBuf>>>) -> Self {
    Server { config, serve_dir }
  }

  pub async fn spawn(
    self,
    notify_sender: Sender<Thread>,
  ) -> Result<(ActixServer, JoinHandle<Result<()>>, ThreadHandle)> {
    let (tx, rx) = mpsc::channel();

    let server_address = self.config.address;
    let serve_dir = Arc::clone(&self.serve_dir);
    let redirect_to_slash = self.config.redirect_to_slash;
    let compression = self.config.compression.clone();

    let root_path_context =
      Arc::new(PathContext::try_from(&self.config).context(CreatePathContext)?);
    let mut path_contexts: Vec<PathContext> = Vec::new();

    if let Some(path_configs) = &self.config.path_configs {
      for path_config in path_configs {
        path_contexts.push(path_config.try_into().context(CreatePathContext)?);
      }
    }

    let path_contexts = Arc::new(path_contexts);

    let (join_handle, thread_handle) = thread::spawn(notify_sender, move || {
      let sys = System::new("http-server");

      let srv = HttpServer::new(move || {
        let mut files_service = Files::new(
          "/",
          Arc::clone(&serve_dir),
          Arc::clone(&root_path_context),
          Arc::clone(&path_contexts),
        );

        if let Some(true) = redirect_to_slash {
          files_service = files_service.redirect_to_slash_directory();
        }

        App::new()
          .wrap(Logger::default())
          .wrap(middleware::Compress::new(match &compression {
            Some(compression) => match compression {
              CompressionConfig::Auto => ContentEncoding::Auto,
              CompressionConfig::Brotli => ContentEncoding::Br,
              CompressionConfig::Disabled => ContentEncoding::Identity,
              CompressionConfig::Deflate => ContentEncoding::Deflate,
              CompressionConfig::Gzip => ContentEncoding::Gzip,
            },
            // Default to auto if not specified in the config.
            _ => ContentEncoding::Auto,
          }))
          .default_service(files_service)
      })
      .bind(server_address)
      .context(Bind)?
      .shutdown_timeout(60)
      .run();

      let _ = tx.send(srv);

      sys.run().context(SystemRun)
    });

    Ok((
      rx.recv().context(ChannelReceive)?,
      join_handle,
      thread_handle,
    ))
  }
}
